#include "parallel/MeshManipulation.h"
#include "Mesh.h"
#include "Traverse.h"
#include "Debug.h"

namespace AMDiS {

  void MeshManipulation::deleteDoubleDofs(std::set<MacroElement*>& newMacroEl)
  {
    std::map<int, MacroElement*> leafInMacroEl;
    
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
    while (elInfo) {
      leafInMacroEl[elInfo->getElement()->getIndex()] = elInfo->getMacroElement();
      elInfo = stack.traverseNext(elInfo);
    }

    deleteDoubleDofs(leafInMacroEl, newMacroEl, 0);
    deleteDoubleDofs(leafInMacroEl, newMacroEl, 1);
  }


  void MeshManipulation::deleteDoubleDofs(std::map<int, MacroElement*>& leafInMacroEl,
					  std::set<MacroElement*>& newMacroEl,
					  int mode)
  {
    FUNCNAME("MeshManipulation::deleteDoubleDofs()");

    std::map<const DegreeOfFreedom*, const DegreeOfFreedom*> mapDelDofs;
    std::map<const DegreeOfFreedom*, int> mapDofsMacro;

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL | Mesh::FILL_NEIGH);
    while (elInfo) {
      Element *el = elInfo->getElement();

      for (int i = 0; i < mesh->getGeo(NEIGH); i++) {

	Element *neigh = elInfo->getNeighbour(i);

	if (!neigh)
	  continue;

	TEST_EXIT_DBG(leafInMacroEl.count(el->getIndex()) == 1)
	  ("Should not happen!\n");
	TEST_EXIT_DBG(leafInMacroEl.count(neigh->getIndex()) == 1)
	  ("No macro element for el %d, that is %d-ith neigbour of element %d!\n",
	   neigh->getIndex(), i, el->getIndex());

	int elMacroIndex = leafInMacroEl[el->getIndex()]->getIndex();
	int neighMacroIndex = leafInMacroEl[neigh->getIndex()]->getIndex();

	if (elMacroIndex == neighMacroIndex)
	  continue;

	if ((mode == 0 && 
	     newMacroEl.count(leafInMacroEl[el->getIndex()]) == 1 &&
	     newMacroEl.count(leafInMacroEl[neigh->getIndex()]) == 1 &&
	     elMacroIndex > neighMacroIndex) ||
	    (mode == 1 &&
	     newMacroEl.count(leafInMacroEl[el->getIndex()]) == 0 &&
	     newMacroEl.count(leafInMacroEl[neigh->getIndex()]) == 1)) {

	  if (el->getEdge(i) != neigh->getEdge(elInfo->getOppVertex(i))) {
	    std::vector<int> elEdgeIndex(2);
	    elEdgeIndex[0] = el->getVertexOfEdge(i, 0);
	    elEdgeIndex[1] = el->getVertexOfEdge(i, 1);

	    std::vector<int> neighEdgeIndex(2);
	    neighEdgeIndex[0] = neigh->getVertexOfEdge(elInfo->getOppVertex(i), 0);
	    neighEdgeIndex[1] = neigh->getVertexOfEdge(elInfo->getOppVertex(i), 1);

	    std::vector<DegreeOfFreedom> elEdgeDof(2);
	    elEdgeDof[0] = el->getDof(elEdgeIndex[0], 0);
	    elEdgeDof[1] = el->getDof(elEdgeIndex[1], 0);

	    std::vector<DegreeOfFreedom> neighEdgeDof(2);
	    neighEdgeDof[0] = neigh->getDof(neighEdgeIndex[0], 0);
	    neighEdgeDof[1] = neigh->getDof(neighEdgeIndex[1], 0);

	    if ((elEdgeDof[0] < elEdgeDof[1] && neighEdgeDof[0] > neighEdgeDof[1]) ||
		(elEdgeDof[0] > elEdgeDof[1] && neighEdgeDof[0] < neighEdgeDof[1])) {
	      std::swap(neighEdgeIndex[0], neighEdgeIndex[1]);
	      std::swap(neighEdgeDof[0], neighEdgeDof[1]);
	    }


	    for (int j = 0; j < 2; j++) {

	      if (neighEdgeDof[j] != elEdgeDof[j]) {
		const DegreeOfFreedom *delDof = neigh->getDof(neighEdgeIndex[j]);
		const DegreeOfFreedom *replaceDof = el->getDof(elEdgeIndex[j]);

		if (mapDelDofs.count(delDof) == 1 && mapDelDofs[delDof] != replaceDof) {
		  if (mapDofsMacro[mapDelDofs[delDof]] > elMacroIndex) {
		    mapDelDofs[replaceDof] = mapDelDofs[delDof];
		  } else {
		    mapDelDofs[mapDelDofs[delDof]] = replaceDof;
		    mapDelDofs[delDof] = replaceDof;
		    mapDofsMacro[replaceDof] = elMacroIndex;
		  }
		} else {
		  mapDelDofs[delDof] = replaceDof;
		  mapDofsMacro[replaceDof] = elMacroIndex;
		}
	      }
	    }	     
	  }
	}
      }


      elInfo = stack.traverseNext(elInfo);
    }


    bool changed = false;
    do {
      changed = false;
      for (std::map<const DegreeOfFreedom*, const DegreeOfFreedom*>::iterator it = mapDelDofs.begin();
	   it != mapDelDofs.end(); ++it) {
	std::map<const DegreeOfFreedom*, const DegreeOfFreedom*>::iterator findIt = mapDelDofs.find(it->second);
	if (findIt != mapDelDofs.end()) {
	  it->second = findIt->second;   
	  changed = true;
	}
      } 
    } while (changed);


    elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      for (int i = 0; i < mesh->getGeo(VERTEX); i++)
	if (mapDelDofs.count(elInfo->getElement()->getDof(i)) == 1)
	  elInfo->getElement()->setDof(i, const_cast<DegreeOfFreedom*>(mapDelDofs[elInfo->getElement()->getDof(i)]));

      elInfo = stack.traverseNext(elInfo);
    }

    for (std::map<const DegreeOfFreedom*, const DegreeOfFreedom*>::iterator it = mapDelDofs.begin();
	 it != mapDelDofs.end(); ++it)
      mesh->freeDof(const_cast<DegreeOfFreedom*>(it->first), VERTEX);
  }

}
